/*#############################################################################
#                                                                             #
# Pakfire - The IPFire package management system                              #
# Copyright (C) 2011 Pakfire development team                                 #
#                                                                             #
# This program is free software: you can redistribute it and/or modify        #
# it under the terms of the GNU General Public License as published by        #
# the Free Software Foundation, either version 3 of the License, or           #
# (at your option) any later version.                                         #
#                                                                             #
# This program is distributed in the hope that it will be useful,             #
# but WITHOUT ANY WARRANTY; without even the implied warranty of              #
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the               #
# GNU General Public License for more details.                                #
#                                                                             #
# You should have received a copy of the GNU General Public License           #
# along with this program.  If not, see <http://www.gnu.org/licenses/>.       #
#                                                                             #
#############################################################################*/

#define PY_SSIZE_T_CLEAN
#include <Python.h>
#include <datetime.h>

#include <errno.h>
#include <stdio.h>

#include <pakfire/filelist.h>
#include <pakfire/package.h>
#include <pakfire/packagelist.h>

#include "file.h"
#include "package.h"
#include "util.h"

PyObject* PyUnicodeList_FromStringArray(char** l) {
	PyObject* list = NULL;
	PyObject* s = NULL;
	int r;

	list = PyList_New(0);
	if (!list)
		goto ERROR;

	for (; l && *l; l++) {
		s = PyUnicode_FromString(*l);
		if (!s)
			goto ERROR;

		r = PyList_Append(list, s);
		Py_DECREF(s);
		if (r)
			goto ERROR;
	}

	return list;

ERROR:
	if (list)
		Py_DECREF(list);

	return NULL;
}

PyObject* PyList_FromPackageList(struct pakfire_packagelist* packagelist) {
	PyObject* list = PyList_New(0);

	const size_t length = pakfire_packagelist_length(packagelist);

	for (size_t i = 0; i < length; i++) {
		struct pakfire_package* package = pakfire_packagelist_get(packagelist, i);

		PyObject* item = new_package(&PackageType, package);
		PyList_Append(list, item);

		pakfire_package_unref(package);
		Py_DECREF(item);
	}

	return list;
}

PyObject* PyList_FromFileList(struct pakfire_filelist* filelist) {
	struct pakfire_file* file = NULL;
	PyObject* f = NULL;
	int r;

	PyObject* list = PyList_New(0);
	if (!list)
		goto ERROR;

	// Fetch size
	const size_t size = pakfire_filelist_length(filelist);

	for (unsigned int i = 0; i < size; i++) {
		file = pakfire_filelist_get(filelist, i);
		if (!file)
			goto ERROR;

		// Create a new File object
		f = new_file(file);
		if (!f)
			goto ERROR;

		// Append the new object to the list
		r = PyList_Append(list, f);
		Py_DECREF(f);

		// If we could not append to the list, we will break
		if (r)
			goto ERROR;

		// Free file
		pakfire_file_unref(file);
	}

	return list;

ERROR:
	Py_XDECREF(list);
	return NULL;
}

PyObject* PyDateTime_FromTime_t(const time_t* t) {
	struct tm buffer;

	struct tm* tm = gmtime_r(t, &buffer);
	if (!tm)
		return NULL;

	PyDateTime_IMPORT;

	return PyDateTime_FromDateAndTime(
		1900 + tm->tm_year,
		1 + tm->tm_mon,
		tm->tm_mday,
		tm->tm_hour,
		tm->tm_min,
		tm->tm_sec,
		0
	);
}

static ssize_t PyObject_AsFileHandle_read(void* cookie, char* buffer, size_t size) {
	PyObject* object = (PyObject*)cookie;
	PyObject* pybuffer = NULL;
	const char* bytes = NULL;
	Py_ssize_t bytes_read = -1;

	// Make sure we acquired the GIL
	PyGILState_STATE state = PyGILState_Ensure();

	// Read data from the object
	pybuffer = PyObject_CallMethod(object, "read", "i", size);
	if (!pybuffer)
		goto ERROR;

	// Handle strings
	if (PyUnicode_Check(pybuffer)) {
		bytes = PyUnicode_AsUTF8AndSize(pybuffer, &bytes_read);
		if (!bytes)
			goto ERROR;

	// Handle bytes
	} else if (PyBytes_Check(pybuffer)) {
		bytes = PyBytes_AsString(pybuffer);
		if (!bytes)
			goto ERROR;

		bytes_read = PyBytes_Size(pybuffer);

	// Fail on anything else
	} else {
		PyErr_SetString(PyExc_TypeError, "read() returned an unexpected data type\n");
		goto ERROR;
	}

	// Copy the data into the target buffer
	memcpy(buffer, bytes, bytes_read);

ERROR:
	if (pybuffer)
		Py_DECREF(pybuffer);

	// Release the GIL
	PyGILState_Release(state);

	return bytes_read;
}

static ssize_t PyObject_AsFileHandle_write(void* cookie, const char* buffer, size_t size) {
	PyObject* object = (PyObject*)cookie;
	ssize_t r = 0;

	// Make sure we acquired the GIL
	PyGILState_STATE state = PyGILState_Ensure();

	PyObject* pybuffer = PyBytes_FromStringAndSize(buffer, size);
	if (!pybuffer)
		goto ERROR;

	PyObject* result = PyObject_CallMethod(object, "write", "b", pybuffer);
	if (!result)
		goto ERROR;

	// It looks like we successfully wrote the data
	r = size;

ERROR:
	Py_XDECREF(pybuffer);

	// Release the GIL
	PyGILState_Release(state);

	return r;
}

static int PyObject_AsFileHandle_close(void* cookie) {
	PyObject* object = (PyObject*)cookie;

	// Make sure we acquired the GIL
	PyGILState_STATE state = PyGILState_Ensure();

	// Decrement the reference counter for the object
	Py_DECREF(object);

	// Release the GIL
	PyGILState_Release(state);

	return 0;
}

static cookie_io_functions_t PyFileHandleMethods = {
	.read = PyObject_AsFileHandle_read,
	.write = PyObject_AsFileHandle_write,
	.close = PyObject_AsFileHandle_close,
};

FILE* PyObject_AsFileHandle(PyObject* object, const char* mode) {
	// We only support reading files
	if (!mode || *mode != 'r') {
		errno = ENOTSUP;
		return NULL;
	}

	// Increment the reference counter for the object
	Py_INCREF(object);

	return fopencookie(object, mode, PyFileHandleMethods);
}
