/*#############################################################################
#                                                                             #
# Pakfire - The IPFire package management system                              #
# Copyright (C) 2017 Pakfire development team                                 #
#                                                                             #
# This program is free software: you can redistribute it and/or modify        #
# it under the terms of the GNU General Public License as published by        #
# the Free Software Foundation, either version 3 of the License, or           #
# (at your option) any later version.                                         #
#                                                                             #
# This program is distributed in the hope that it will be useful,             #
# but WITHOUT ANY WARRANTY; without even the implied warranty of              #
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the               #
# GNU General Public License for more details.                                #
#                                                                             #
# You should have received a copy of the GNU General Public License           #
# along with this program.  If not, see <http://www.gnu.org/licenses/>.       #
#                                                                             #
#############################################################################*/

#include <errno.h>
#include <linux/limits.h>
#include <stdlib.h>
#include <string.h>
#include <sys/mman.h>
#include <sys/stat.h>
#include <unistd.h>

// OpenSSL
#include <openssl/err.h>
#include <openssl/evp.h>
#include <openssl/pem.h>
#include <openssl/rand.h>

#include <pakfire/constants.h>
#include <pakfire/i18n.h>
#include <pakfire/key.h>
#include <pakfire/logging.h>
#include <pakfire/pakfire.h>
#include <pakfire/private.h>
#include <pakfire/string.h>
#include <pakfire/util.h>

// The maximum length of the comment
#define COMMENT_MAX		1024

// Size of the buffer to allocate for error messages
#define ERROR_MAX		1024

struct pakfire_key {
	struct pakfire* pakfire;
	int nrefs;

	// Algorithm
	pakfire_key_algo_t algo;

	// Key ID
	pakfire_key_id id;

	// The public (or private) key
	EVP_PKEY* pkey;

	// Comment
	char comment[COMMENT_MAX];
};

struct pakfire_key_public_key {
	unsigned char sig_algo[2];
	pakfire_key_id id;
	unsigned char pubkey[32];
};

struct pakfire_key_private_key {
	unsigned char sig_algo[2];
    unsigned char kdf_algo[2];
	uint32_t kdf_rounds;
    unsigned char kdf_salt[16];
	unsigned char checksum[8];
	pakfire_key_id id;
	struct {
		unsigned char secret[32];
		unsigned char public[32];
	} keys;
};

struct pakfire_key_signature {
	unsigned char sig_algo[2];
	pakfire_key_id key_id;
	unsigned char signature[64];
};

static int pakfire_key_id_equals(const pakfire_key_id* id1, const pakfire_key_id* id2) {
	return !memcmp(*id1, *id2, sizeof(*id1));
}

static int pakfire_key_create(struct pakfire_key** key, struct pakfire* pakfire,
		const pakfire_key_algo_t algo, const pakfire_key_id id, EVP_PKEY* pkey, const char* comment) {
	int r;

	if (!pkey) {
		errno = EINVAL;
		return 1;
	}

	// Allocate memory
	struct pakfire_key* k = calloc(1, sizeof(*k));
	if (!k)
		return 1;

	// Initialize pakfire and reference counter
	k->pakfire = pakfire_ref(pakfire);
	k->nrefs = 1;

	// Store the algorithm
	switch (algo) {
		case PAKFIRE_KEY_ALGO_ED25519:
			k->algo = algo;
			break;

		default:
			ERROR(pakfire, "Unsupported key algorithm %d\n", algo);
			errno = ENOTSUP;
			goto ERROR;
	}

	// Store the key ID
	memcpy(k->id, id, sizeof(k->id));

	// Keep a reference to this key
	EVP_PKEY_up_ref(pkey);
	k->pkey = pkey;

	// Store the comment
	if (comment) {
		r = pakfire_string_set(k->comment, comment);
		if (r)
			goto ERROR;

		// Remove any trailing newline
		pakfire_remove_trailing_newline(k->comment);
	}

	*key = k;
	return 0;

ERROR:
	pakfire_key_unref(k);

	return 1;
}

static void pakfire_key_free(struct pakfire_key* key) {
	// Free the key
	if (key->pkey)
		EVP_PKEY_free(key->pkey);

	pakfire_unref(key->pakfire);
	free(key);
}

PAKFIRE_EXPORT struct pakfire_key* pakfire_key_ref(struct pakfire_key* key) {
	++key->nrefs;

	return key;
}

PAKFIRE_EXPORT void pakfire_key_unref(struct pakfire_key* key) {
	if (--key->nrefs > 0)
		return;

	pakfire_key_free(key);
}

PAKFIRE_EXPORT pakfire_key_id* pakfire_key_get_id(struct pakfire_key* key) {
	return &key->id;
}

PAKFIRE_EXPORT const char* pakfire_key_get_algo(struct pakfire_key* key) {
	switch (key->algo) {
		case PAKFIRE_KEY_ALGO_ED25519:
			return "Ed255919";

		case PAKFIRE_KEY_ALGO_NULL:
			break;
	}

	return NULL;
}

PAKFIRE_EXPORT const char* pakfire_key_get_comment(struct pakfire_key* key) {
	return key->comment;
}

PAKFIRE_EXPORT int pakfire_key_generate(struct pakfire_key** key, struct pakfire* pakfire,
		const pakfire_key_algo_t algo, const char* comment) {
	EVP_PKEY* pkey = NULL;
	EVP_PKEY_CTX* pctx = NULL;
	pakfire_key_id key_id;
	char error[ERROR_MAX];
	int id;
	int r;

	switch (algo) {
		case PAKFIRE_KEY_ALGO_ED25519:
			id = EVP_PKEY_ED25519;
			break;

		default:
			ERROR(pakfire, "Invalid key algorithm %d\n", algo);
			errno = EINVAL;
			return 1;
	}

	// Generate a random key ID
	r = RAND_bytes((unsigned char*)&key_id, sizeof(key_id));
	if (r < 0) {
		ERROR(pakfire, "Could not generate the key ID\n");
		r = 1;
		goto ERROR;
	}

	// Setup the context
	pctx = EVP_PKEY_CTX_new_id(id, NULL);
	if (!pctx) {
		ERROR(pakfire, "Could not allocate the OpenSSL context: %m\n");
		r = 1;
		goto ERROR;
	}

	// Prepare the context for key generation
	r = EVP_PKEY_keygen_init(pctx);
	if (r < 1) {
		ERR_error_string_n(r, error, sizeof(error));

		ERROR(pakfire, "Could not prepare the context: %s\n", error);
		r = 1;
		goto ERROR;
	}

	// Generate the key
	r = EVP_PKEY_keygen(pctx, &pkey);
	if (r < 1) {
		ERR_error_string_n(r, error, sizeof(error));

		ERROR(pakfire, "Could not generate the key: %s\n", error);
		r = 1;
		goto ERROR;
	}

	// Create a key object
	r = pakfire_key_create(key, pakfire, algo, key_id, pkey, comment);
	if (r)
		goto ERROR;

	// Success
	r = 0;

ERROR:
	if (pctx)
		EVP_PKEY_CTX_free(pctx);
	if (pkey)
		EVP_PKEY_free(pkey);

	return r;
}

/*
	Import
*/

static int pakfire_key_import_secret_key(struct pakfire_key** key,
		struct pakfire* pakfire, const char* comment,
		const struct pakfire_key_private_key* buffer) {
	const pakfire_key_algo_t algo = PAKFIRE_KEY_ALGO_ED25519;
	EVP_PKEY* pkey = NULL;
	char error[ERROR_MAX];
	int r;

	unsigned char checksum[64];
	unsigned int length = sizeof(checksum);

	// Check for input
	if (!buffer) {
		errno = EINVAL;
		return 1;
	}

	// Check the signature algorithm
	if (buffer->sig_algo[0] != 'E' || buffer->sig_algo[1] != 'd') {
		ERROR(pakfire, "Unsupported signature algorithm\n");
		errno = ENOTSUP;
		r = 1;
		goto ERROR;
	}

	// Check the KDF algorithm
	if (buffer->kdf_algo[0] != 'B' || buffer->kdf_algo[1] != 'K') {
		ERROR(pakfire, "Unsupported KDF algorithm\n");
		errno = ENOTSUP;
		r = 1;
		goto ERROR;
	}

	// We don't support encrypted keys here
	if (buffer->kdf_rounds) {
		ERROR(pakfire, "Encrypted keys are not supported\n");
		errno = ENOTSUP;
		r = 1;
		goto ERROR;
	}

	// Compute a SHA512 checksum over the key material
	r = EVP_Digest(&buffer->keys, sizeof(buffer->keys), checksum, &length, EVP_sha512(), NULL);
	if (r < 0) {
		ERROR(pakfire, "Could not compute the checksum: %m\n");
		r = 1;
		goto ERROR;
	}

	// Compare the checksum
	if (memcmp(buffer->checksum, checksum, sizeof(buffer->checksum)) != 0) {
		ERROR(pakfire, "Checksum mismatch\n");
		r = 1;
		goto ERROR;
	}

	// Load the private key
	pkey = EVP_PKEY_new_raw_private_key(EVP_PKEY_ED25519, NULL,
		buffer->keys.secret, sizeof(buffer->keys.secret));
	if (!pkey) {
		ERR_error_string_n(ERR_get_error(), error, sizeof(error));

		ERROR(pakfire, "Could not load secret key: %s\n", error);
		r = 1;
		goto ERROR;
	}

	// Create a new key object
	r = pakfire_key_create(key, pakfire, algo, buffer->id, pkey, comment);
	if (r)
		goto ERROR;

ERROR:
	if (pkey)
		EVP_PKEY_free(pkey);

	return 0;
}

static int pakfire_key_import_public_key(struct pakfire_key** key,
		struct pakfire* pakfire, const char* comment,
		const struct pakfire_key_public_key* buffer) {
	const pakfire_key_algo_t algo = PAKFIRE_KEY_ALGO_ED25519;
	EVP_PKEY* pkey = NULL;
	char error[ERROR_MAX];
	int r;

	// Check for input
	if (!buffer) {
		errno = EINVAL;
		return 1;
	}

	// Check the signature algorithm
	if (buffer->sig_algo[0] != 'E' || buffer->sig_algo[1] != 'd') {
		ERROR(pakfire, "Unsupported signature algorithm\n");
		errno = ENOTSUP;
		r = 1;
		goto ERROR;
	}

	// Load the public key
	pkey = EVP_PKEY_new_raw_public_key(EVP_PKEY_ED25519, NULL,
		buffer->pubkey, sizeof(buffer->pubkey));
	if (!pkey) {
		ERR_error_string_n(ERR_get_error(), error, sizeof(error));

		ERROR(pakfire, "Could not load public key: %s\n", error);
		r = 1;
		goto ERROR;
	}

	// Create a new key object
	r = pakfire_key_create(key, pakfire, algo, buffer->id, pkey, comment);
	if (r)
		goto ERROR;

ERROR:
	if (pkey)
		EVP_PKEY_free(pkey);

	return r;
}

PAKFIRE_EXPORT int pakfire_key_import(struct pakfire_key** key,
		struct pakfire* pakfire, FILE* f) {
	void* buffer = NULL;
	size_t buffer_length = 0;
	int r;

	char* line = NULL;
	size_t length = 0;
	size_t lineno = 0;

	char comment[COMMENT_MAX];

	for (;;) {
		ssize_t bytes_read = getline(&line, &length, f);
		if (bytes_read < 0)
			break;

		// Increment the line counter
		lineno++;

		switch (lineno) {
			// The first line must start with "untrusted comment:"
			case 1:
				if (!pakfire_string_startswith(line, "untrusted comment: ")) {
					ERROR(pakfire, "The first line must start with 'untrusted comment: '"
						" and not %s\n", line);
					errno = EINVAL;
					r = 1;
					goto ERROR;
				}

				// Copy the comment
				r = pakfire_string_set(comment, line + strlen("untrusted comment: "));
				if (r) {
					ERROR(pakfire, "Could not copy comment: %m\n");
					r = 1;
					goto ERROR;
				}

				break;

			// The second line should hold the key
			case 2:
				// Decode the key
				r = pakfire_b64decode(pakfire, &buffer, &buffer_length, line);
				if (r) {
					ERROR(pakfire, "Could not decode the key: %m\n");
					errno = EINVAL;
					r = 1;
					goto ERROR;
				}

				// What kind of key do we have?
				switch (buffer_length) {
					// Public Key
					case sizeof(struct pakfire_key_public_key):
						r = pakfire_key_import_public_key(key, pakfire, comment,
							(struct pakfire_key_public_key*)buffer);
						break;

					// Private Key
					case sizeof(struct pakfire_key_private_key):
						r = pakfire_key_import_secret_key(key, pakfire, comment,
							(struct pakfire_key_private_key*)buffer);
						break;

					// Unknown key
					default:
						ERROR(pakfire, "Unsupported key type\n");
						errno = ENOTSUP;
						r = 1;
						goto ERROR;
				}
				break;

			// Ignore any further data
			default:
				break;
		}
	}

	// Success!
	r = 0;

ERROR:
	if (buffer) {
		OPENSSL_cleanse(buffer, buffer_length);
		free(buffer);
	}
	if (line)
		free(line);

	return r;
}

int pakfire_key_import_from_string(struct pakfire_key** key,
		struct pakfire* pakfire, const char* data, const size_t length) {
	FILE* f = NULL;
	int r;

	// Map the data to a file
	f = fmemopen((char*)data, length, "r");
	if (!f) {
		ERROR(pakfire, "Could not map the key to file: %m\n");
		r = 1;
		goto ERROR;
	}

	// Import the key
	r = pakfire_key_import(key, pakfire, f);

ERROR:
	if (f)
		fclose(f);

	return r;
}

static int pakfire_key_get_public_key(struct pakfire_key* key,
		unsigned char* buffer, const size_t length) {
	char error[ERROR_MAX];
	int r;

	size_t l = length;

	// Extract the raw public key
	r = EVP_PKEY_get_raw_public_key(key->pkey, buffer, &l);
	if (r < 0) {
		ERR_error_string_n(ERR_get_error(), error, sizeof(error));

		ERROR(key->pakfire, "Could not extract the public key: %s\n", error);
		return 1;
	}

	if (l > length) {
		ERROR(key->pakfire, "The buffer was too small to write the public key\n");
		errno = ENOBUFS;
		return 1;
	}

	return 0;
}

static int pakfire_key_get_secret_key(struct pakfire_key* key,
		unsigned char* buffer, const size_t length) {
	char error[ERROR_MAX];
	int r;

	size_t l = length;

	// Extract the raw secret key
	r = EVP_PKEY_get_raw_private_key(key->pkey, buffer, &l);
	if (r < 0) {
		ERR_error_string_n(ERR_get_error(), error, sizeof(error));

		ERROR(key->pakfire, "Could not extract the secret key: %s\n", error);
		return 1;
	}

	if (l > length) {
		ERROR(key->pakfire, "The buffer was too small to write the secret key\n");
		errno = ENOBUFS;
		return 1;
	}

	return 0;
}

/*
	Export
*/

static int pakfire_key_export_private_key(struct pakfire_key* key,
		struct pakfire_key_private_key* buffer) {
	unsigned char checksum[64];
	unsigned int length = sizeof(checksum);
	int r;

	// Write the algorithm
	switch (key->algo) {
		case PAKFIRE_KEY_ALGO_ED25519:
			// Signature Algortihm
			buffer->sig_algo[0] = 'E';
			buffer->sig_algo[1] = 'd';

			// KDF Algorithm
			buffer->kdf_algo[0] = 'B';
			buffer->kdf_algo[1] = 'K';
			break;

		default:
			ERROR(key->pakfire, "Unknown algorithm\n");
			return 1;
	}

	// Generate a salt
	r = RAND_bytes(buffer->kdf_salt, sizeof(buffer->kdf_salt));
	if (r < 1) {
		ERROR(key->pakfire, "Could not generate salt\n");
		return 1;
	}

	// Copy the key ID
	memcpy(buffer->id, key->id, sizeof(buffer->id));

	// Write the public key
	r = pakfire_key_get_public_key(key, buffer->keys.public, sizeof(buffer->keys.public));
	if (r) {
		ERROR(key->pakfire, "Could not export the public key: %m\n");
		return r;
	}

	// Write the secret key
	r = pakfire_key_get_secret_key(key, buffer->keys.secret, sizeof(buffer->keys.secret));
	if (r) {
		ERROR(key->pakfire, "Could not export the secret key: %m\n");
		return r;
	}

	// Compute a SHA512 checksum over the key material
	r = EVP_Digest(&buffer->keys, sizeof(buffer->keys), checksum, &length, EVP_sha512(), NULL);
	if (r < 0) {
		ERROR(key->pakfire, "Could not compute the checksum: %m\n");
		return 1;
	}

	// Copy the first couple of bytes of the checksum
	memcpy(buffer->checksum, checksum, sizeof(buffer->checksum));

	// Success!
	return 0;
}

static int pakfire_key_export_public_key(struct pakfire_key* key,
		struct pakfire_key_public_key* buffer) {
	int r;

	// Write the algorithm
	switch (key->algo) {
		case PAKFIRE_KEY_ALGO_ED25519:
			buffer->sig_algo[0] = 'E';
			buffer->sig_algo[1] = 'd';
			break;

		default:
			ERROR(key->pakfire, "Unknown algorithm\n");
			return 1;
	}

	// Copy the key ID
	memcpy(buffer->id, key->id, sizeof(buffer->id));

	// Write the public key
	r = pakfire_key_get_public_key(key, buffer->pubkey, sizeof(buffer->pubkey));
	if (r) {
		ERROR(key->pakfire, "Could not export the public key: %m\n");
		return r;
	}

	// Success!
	return 0;
}

PAKFIRE_EXPORT int pakfire_key_export(struct pakfire_key* key, FILE* f,
		const pakfire_key_export_mode_t mode) {
	struct pakfire_key_public_key public_key = { 0 };
	struct pakfire_key_private_key private_key = { 0 };
	int r;

	BIO* bio = NULL;
	BIO* b64 = NULL;

	// Write the comment
	if (*key->comment) {
		r = fprintf(f, "untrusted comment: %s\n", key->comment);
		if (r < 0) {
			ERROR(key->pakfire, "Could not write comment: %m\n");
			r = 1;
			goto ERROR;
		}
	}

	// Setup the base64 encoder
	b64 = BIO_new(BIO_f_base64());
	if (!b64) {
		ERROR(key->pakfire, "Could not setup the base64 encoder\n");
		r = 1;
		goto ERROR;
	}

	// Disable line breaks and a trailing newline
	BIO_set_flags(b64, BIO_FLAGS_BASE64_NO_NL);

	bio = BIO_new_fp(f, BIO_NOCLOSE);
	if (!bio) {
		ERROR(key->pakfire, "Could not open BIO\n");
		r = 1;
		goto ERROR;
	}

	BIO_push(b64, bio);

	switch (mode) {
		case PAKFIRE_KEY_EXPORT_MODE_PUBLIC:
			r = pakfire_key_export_public_key(key, &public_key);
			if (r)
				goto ERROR;

			// Write the output
			r = BIO_write(b64, &public_key, sizeof(public_key));
			if (r < 0) {
				ERROR(key->pakfire, "Could not write the public key\n");
				r = 1;
				goto ERROR;
			}
			break;

		case PAKFIRE_KEY_EXPORT_MODE_PRIVATE:
			r = pakfire_key_export_private_key(key, &private_key);
			if (r)
				goto ERROR;

			// Write the output
			r = BIO_write(b64, &private_key, sizeof(private_key));
			if (r < 0) {
				ERROR(key->pakfire, "Could not write the private key\n");
				r = 1;
				goto ERROR;
			}
			break;

		default:
			errno = EINVAL;
			r = 1;
			goto ERROR;
	}

	// Flush all buffers
	BIO_flush(b64);

	// Append a trailing newline
	fprintf(f, "\n");

	// Success!
	r = 0;

ERROR:
	if (b64)
		BIO_free(b64);
	if (bio)
		BIO_free(bio);

	return r;
}

PAKFIRE_EXPORT char* pakfire_key_dump(struct pakfire_key* key) {
	char* buffer = NULL;
	FILE* f = NULL;
	int fd = -1;
	int r;

	// Allocate a buffer in memory
	fd = memfd_create("pakfire-key-dump", MFD_CLOEXEC);
	if (fd < 0) {
		ERROR(key->pakfire, "Could not allocate a temporary file: %m\n");
		r = 1;
		goto ERROR;
	}

	// Re-open as FILE handle
	f = fdopen(fd, "r+");
	if (!f) {
		ERROR(key->pakfire, "Could not open file handle for temporary file: %m\n");
		r = 1;
		goto ERROR;
	}

	// Export the public part of the key
	r = pakfire_key_export(key, f, PAKFIRE_KEY_EXPORT_MODE_PUBLIC);
	if (r) {
		ERROR(key->pakfire, "Could not export key: %m\n");
		goto ERROR;
	}

	// Determine the length of the buffer
	size_t length = ftell(f);

	// Go back to the beginning of the file
	rewind(f);

	// Allocate a buffer
	buffer = calloc(1, length + 1);
	if (!buffer) {
		ERROR(key->pakfire, "Could not allocate buffer of %zu byte(s)\n", length + 1);
		r = 1;
		goto ERROR;
	}

	// Read everything into the buffer
	size_t bytes_read = fread(buffer, 1, length, f);
	if (bytes_read < length) {
		ERROR(key->pakfire, "Could not read back the buffer: %m\n");
		r = 1;
		goto ERROR;
	}

	// Success!
	r = 0;

ERROR:
	if (f)
		fclose(f);
	if (fd >= 0)
		close(fd);

	// Free the buffer on error
	if (r && buffer) {
		free(buffer);
		buffer = NULL;
	}

	return buffer;
}

/*
	Sign
*/

static int __pakfire_key_sign(struct pakfire_key* key,
		struct pakfire_key_signature* signature, const void* data, const size_t length) {
	EVP_MD_CTX* mdctx = NULL;
	char error[ERROR_MAX];
	int r;

	// Check inputs
	if (!signature || !data || !length) {
		errno = EINVAL;
		return 1;
	}

	// Set algorithm
	signature->sig_algo[0] = 'E';
	signature->sig_algo[1] = 'd';

	// Set the key ID
	memcpy(signature->key_id, key->id, sizeof(signature->key_id));

	// Create a message digest context
	mdctx = EVP_MD_CTX_new();
	if (!mdctx) {
		ERROR(key->pakfire, "Could not initialize the message digest context: %m\n");
		r = 1;
		goto ERROR;
	}

	// Setup the context
	r = EVP_DigestSignInit(mdctx, NULL, NULL, NULL, key->pkey);
	if (r < 1) {
		ERROR(key->pakfire, "Could not setup context\n");
		r = 1;
		goto ERROR;
	}

	size_t signature_length = sizeof(signature->signature);

	// Create a signature over the digest
	r = EVP_DigestSign(mdctx, signature->signature, &signature_length, data, length);
	if (r < 1) {
		ERR_error_string_n(ERR_get_error(), error, sizeof(error));

		ERROR(key->pakfire, "Could not sign content: %s\n", error);
		r = 1;
		goto ERROR;
	}

	// Success!
	r = 0;

ERROR:
	if (mdctx)
		EVP_MD_CTX_free(mdctx);

	return r;
}

PAKFIRE_EXPORT int pakfire_key_sign(struct pakfire_key* key,
		FILE* f, const void* data, const size_t length, const char* comment) {
	struct pakfire_key_signature signature = { 0 };
	char* s = NULL;
	int r;

	// Create a signature
	r = __pakfire_key_sign(key, &signature, data, length);
	if (r)
		return r;

	// Write the comment
	if (comment) {
		r = fprintf(f, "untrusted comment: %s\n", comment);
		if (r < 0) {
			ERROR(key->pakfire, "Could not write comment: %m\n");
			r = 1;
			goto ERROR;
		}
	}

	// Encode the signature to base64
	r = pakfire_b64encode(key->pakfire, &s, &signature, sizeof(signature));
	if (r)
		goto ERROR;

	// Write the signature
	r = fprintf(f, "%s\n", s);
	if (r < 0) {
		ERROR(key->pakfire, "Could not write the signature: %m\n");
		r = 1;
		goto ERROR;
	}

	// Ensure the entire signature is being written
	fflush(f);

	// Success!
	r = 0;

ERROR:
	if (s)
		free(s);

	return r;
}

int pakfire_key_signf(struct pakfire_key* key, FILE* s, FILE* f, const char* comment) {
	char* buffer = NULL;
	size_t length = 0;
	int r;

	// Load the entire content into memory
	r = pakfire_read_file_into_buffer(f, &buffer, &length);
	if (r)
		goto ERROR;

	// Sign!
	r = pakfire_key_sign(key, s, buffer, length, comment);

ERROR:
	if (buffer)
		free(buffer);

	return r;
}

static int pakfire_key_read_signature(struct pakfire_key* key,
		struct pakfire_key_signature* signature, FILE* f) {
	void* buffer = NULL;
	size_t buffer_length = 0;
	int r;

	char* line = NULL;
	size_t length = 0;
	size_t lineno = 0;

	for (;;) {
		ssize_t bytes_read = getline(&line, &length, f);
		if (bytes_read < 0)
			break;

		// Increment the line counter
		lineno++;

		switch (lineno) {
			// The first line must start with "untrusted comment:"
			case 1:
				if (!pakfire_string_startswith(line, "untrusted comment:")) {
					ERROR(key->pakfire, "The first line must start with 'untrusted comment:'\n");
					errno = EINVAL;
					r = 1;
					goto ERROR;
				}
				break;

			// The second line should hold the signature
			case 2:
				// Decode the key
				r = pakfire_b64decode(key->pakfire, &buffer, &buffer_length, line);
				if (r) {
					ERROR(key->pakfire, "Could not decode the signature: %m\n");
					errno = EINVAL;
					r = 1;
					goto ERROR;
				}

				// What kind of signature do we have?
				switch (buffer_length) {
					case sizeof(*signature):
						// Copy the buffer to the signature
						memcpy(signature, buffer, sizeof(*signature));

						// Check if we support the signature type
						if (signature->sig_algo[0] != 'E' || signature->sig_algo[1] != 'd') {
							ERROR(key->pakfire, "Unknown signature type\n");
							errno = ENOTSUP;
							r = 1;
							goto ERROR;
						}
						break;

					default:
						ERROR(key->pakfire, "Unknown signature type\n");
						errno = ENOTSUP;
						r = 1;
						goto ERROR;
				}
				break;

			// Ignore any further data
			default:
				break;
		}
	}

ERROR:
	if (buffer)
		free(buffer);
	if (line)
		free(line);

	return r;
}

static int pakfire_key_verify_signature(struct pakfire_key* key,
		const struct pakfire_key_signature* signature, const void* data, const size_t length) {
	EVP_MD_CTX* mdctx = NULL;
	int r;

	DEBUG(key->pakfire, "Verifying signature...\n");

	// Check the key ID
	if (!pakfire_key_id_equals(&key->id, &signature->key_id)) {
		ERROR(key->pakfire, "The signature has been created with a different key\n");
		errno = EBADMSG;
		r = 1;
		goto ERROR;
	}

	// Create message digest context
	mdctx = EVP_MD_CTX_new();
	if (!mdctx) {
		ERROR(key->pakfire, "Could not create the message digest context\n");
		r = 1;
		goto ERROR;
	}

	// Setup the context for verification
	r = EVP_DigestVerifyInit(mdctx, NULL, NULL, NULL, key->pkey);
	if (r < 1) {
		ERROR(key->pakfire, "Could not setup the verification context\n");
		r = 1;
		goto ERROR;
	}

	// Verify the signature
	r = EVP_DigestVerify(mdctx, signature->signature, sizeof(signature->signature),
			data, length);
	switch (r) {
		// Fail
		case 0:
			ERROR(key->pakfire, "Signature verification failed\n");
			errno = EBADMSG;
			r = 1;
			break;

		// Success
		case 1:
			DEBUG(key->pakfire, "Signature verification successful\n");
			r = 0;
			break;

		// Error
		default:
			ERROR(key->pakfire, "Could not perform signature verification\n");
			r = 1;
			goto ERROR;
	}

ERROR:
	if (mdctx)
		EVP_MD_CTX_free(mdctx);

	return r;
}

PAKFIRE_EXPORT int pakfire_key_verify(struct pakfire_key* key, FILE* f,
		const void* data, const size_t length) {
	struct pakfire_key_signature signature = { 0 };
	int r;

	// Read the signature
	r = pakfire_key_read_signature(key, &signature, f);
	if (r) {
		ERROR(key->pakfire, "Could not read signature: %m\n");
		return r;
	}

	// Verify signature
	r = pakfire_key_verify_signature(key, &signature, data, length);
	if (r) {
		ERROR(key->pakfire, "Could not verify signature: %m\n");
		return r;
	}

	return 0;
}
